@@ -44,6 +44,8 @@ const (
 	deviceScopeConstraintsFeature = "device-scope-constraints"
 	// feature label for plug-names/slot-names constraints
 	nameConstraintsFeature = "name-constraints"
+	// feature label for alt attribute matcher usage
+	altAttrMatcherFeature = "alt-attr-matcher"
 )
 
 type attrMatcher interface {
@@ -293,6 +295,9 @@ func compileAltAttrMatcher(cc compileContext, l []interface{}) (attrMatcher, err
 }
 
 func (matcher altAttrMatcher) feature(flabel string) bool {
+	if flabel == altAttrMatcherFeature {
+		return true
+	}
 	for _, alt := range matcher.alts {
 		if alt.feature(flabel) {
 			return true
@@ -302,6 +307,14 @@ func (matcher altAttrMatcher) feature(flabel string) bool {
 }
 
 func (matcher altAttrMatcher) match(apath string, v interface{}, ctx AttrMatchContext) error {
+	// if the value is a list apply the alternative matcher to each element
+	// like we do for other matchers
+	switch x := v.(type) {
+	case []interface{}:
+		return matchList(apath, matcher, x, ctx)
+	default:
+	}
+
 	var firstErr error
 	for _, alt := range matcher.alts {
 		err := alt.match(apath, v, ctx)
@@ -517,7 +530,7 @@ type nameMatcher interface {
 
 var (
 	// validates special name constraints like $INTERFACE
-	validSpecialNameConstraint = regexp.MustCompile("^\\$[A-Z][A-Z0-9_]*$")
+	validSpecialNameConstraint = regexp.MustCompile(`^\$[A-Z][A-Z0-9_]*$`)
 )
 
 func compileNameMatcher(whichName string, v interface{}) (nameMatcher, error) {
@@ -599,9 +612,9 @@ func (nc *NameConstraints) Check(whichName, name string, special map[string]stri
 // rules
 
 var (
-	validSnapType  = regexp.MustCompile("^(?:core|kernel|gadget|app)$")
-	validDistro    = regexp.MustCompile("^[-0-9a-z._]+$")
-	validPublisher = regexp.MustCompile("^(?:[a-z0-9A-Z]{32}|[-a-z0-9]{2,28}|\\$[A-Z][A-Z0-9_]*)$") // account ids look like snap-ids or are nice identifiers, support our own special markers $MARKER
+	validSnapType  = regexp.MustCompile(`^(?:core|kernel|gadget|app)$`)
+	validDistro    = regexp.MustCompile(`^[-0-9a-z._]+$`)
+	validPublisher = regexp.MustCompile(`^(?:[a-z0-9A-Z]{32}|[-a-z0-9]{2,28}|\$[A-Z][A-Z0-9_]*)$`) // account ids look like snap-ids or are nice identifiers, support our own special markers $MARKER
 
 	validIDConstraints = map[string]*regexp.Regexp{
 		"slot-snap-type":    validSnapType,
