@@ -29,6 +29,7 @@ import (
 	"github.com/snapcore/snapd/asserts/assertstest"
 	"github.com/snapcore/snapd/asserts/snapasserts"
 	"github.com/snapcore/snapd/snap"
+	"github.com/snapcore/snapd/snap/naming"
 )
 
 type validationSetsSuite struct{}
@@ -271,7 +272,7 @@ func (s *validationSetsSuite) TestCheckInstalledSnapsNoValidationSets(c *C) {
 	snaps := []*snapasserts.InstalledSnap{
 		snapasserts.NewInstalledSnap("snap-a", "mysnapaaaaaaaaaaaaaaaaaaaaaaaaaa", snap.R(1)),
 	}
-	err := valsets.CheckInstalledSnaps(snaps)
+	err := valsets.CheckInstalledSnaps(snaps, nil)
 	c.Assert(err, IsNil)
 }
 
@@ -551,7 +552,7 @@ func (s *validationSetsSuite) TestCheckInstalledSnaps(c *C) {
 	}
 
 	for i, tc := range tests {
-		err := valsets.CheckInstalledSnaps(tc.snaps)
+		err := valsets.CheckInstalledSnaps(tc.snaps, nil)
 		if err == nil {
 			c.Assert(tc.expectedInvalid, IsNil)
 			c.Assert(tc.expectedMissing, IsNil)
@@ -567,6 +568,59 @@ func (s *validationSetsSuite) TestCheckInstalledSnaps(c *C) {
 	}
 }
 
+func (s *validationSetsSuite) TestCheckInstalledSnapsIgnoreValidation(c *C) {
+	// require: snapB rev 3, snapC rev 2.
+	// invalid: snapA
+	vs := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "acme",
+		"series":       "16",
+		"account-id":   "acme",
+		"name":         "fooname",
+		"sequence":     "1",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "snap-a",
+				"id":       "mysnapaaaaaaaaaaaaaaaaaaaaaaaaaa",
+				"presence": "invalid",
+			},
+			map[string]interface{}{
+				"name":     "snap-b",
+				"id":       "mysnapbbbbbbbbbbbbbbbbbbbbbbbbbb",
+				"revision": "3",
+				"presence": "required",
+			},
+			map[string]interface{}{
+				"name":     "snap-c",
+				"id":       "mysnapcccccccccccccccccccccccccc",
+				"revision": "2",
+				"presence": "optional",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	valsets := snapasserts.NewValidationSets()
+	c.Assert(valsets.Add(vs), IsNil)
+
+	snapA := snapasserts.NewInstalledSnap("snap-a", "mysnapaaaaaaaaaaaaaaaaaaaaaaaaaa", snap.R(1))
+	snapB := snapasserts.NewInstalledSnap("snap-b", "mysnapbbbbbbbbbbbbbbbbbbbbbbbbbb", snap.R(3))
+	snapBinvRev := snapasserts.NewInstalledSnap("snap-b", "mysnapbbbbbbbbbbbbbbbbbbbbbbbbbb", snap.R(8))
+
+	// sanity check
+	c.Check(valsets.CheckInstalledSnaps([]*snapasserts.InstalledSnap{snapA, snapB}, nil), ErrorMatches, "validation sets assertions are not met:\n"+
+		"- invalid snaps:\n"+
+		"  - snap-a \\(invalid for sets acme/fooname\\)")
+	// snapA is invalid but ignore-validation is set so it's ok
+	c.Check(valsets.CheckInstalledSnaps([]*snapasserts.InstalledSnap{snapA, snapB}, map[string]bool{"snap-a": true}), IsNil)
+
+	// sanity check
+	c.Check(valsets.CheckInstalledSnaps([]*snapasserts.InstalledSnap{snapBinvRev}, nil), ErrorMatches, "validation sets assertions are not met:\n"+
+		"- snaps at wrong revisions:\n"+
+		"  - snap-b \\(required at revision 3 by sets acme/fooname\\)")
+	// snapB is at the wrong revision, but ignore-validation is set so it's ok
+	c.Check(valsets.CheckInstalledSnaps([]*snapasserts.InstalledSnap{snapBinvRev}, map[string]bool{"snap-b": true}), IsNil)
+}
+
 func (s *validationSetsSuite) TestCheckInstalledSnapsErrorFormat(c *C) {
 	vs1 := assertstest.FakeAssertion(map[string]interface{}{
 		"type":         "validation-set",
@@ -640,7 +694,7 @@ func (s *validationSetsSuite) TestCheckInstalledSnapsErrorFormat(c *C) {
 	}
 
 	for i, tc := range tests {
-		err := valsets.CheckInstalledSnaps(tc.snaps)
+		err := valsets.CheckInstalledSnaps(tc.snaps, nil)
 		c.Assert(err, NotNil, Commentf("#%d", i))
 		c.Assert(err, ErrorMatches, tc.errorMsg, Commentf("#%d: ", i))
 	}
@@ -652,3 +706,193 @@ func (s *validationSetsSuite) TestSortByRevision(c *C) {
 	sort.Sort(snapasserts.ByRevision(revs))
 	c.Assert(revs, DeepEquals, []snap.Revision{snap.R(-1), snap.R(4), snap.R(5), snap.R(10)})
 }
+
+func (s *validationSetsSuite) TestCheckPresenceRequired(c *C) {
+	valset1 := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "account-id",
+		"series":       "16",
+		"account-id":   "account-id",
+		"name":         "my-snap-ctl",
+		"sequence":     "1",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "my-snap",
+				"id":       "mysnapididididididididididididid",
+				"presence": "required",
+				"revision": "7",
+			},
+			map[string]interface{}{
+				"name":     "other-snap",
+				"id":       "123456ididididididididididididid",
+				"presence": "optional",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	valset2 := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "account-id",
+		"series":       "16",
+		"account-id":   "account-id",
+		"name":         "my-snap-ctl2",
+		"sequence":     "2",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "my-snap",
+				"id":       "mysnapididididididididididididid",
+				"presence": "required",
+				"revision": "7",
+			},
+			map[string]interface{}{
+				"name":     "other-snap",
+				"id":       "123456ididididididididididididid",
+				"presence": "invalid",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	// my-snap required but no specific revision set.
+	valset3 := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "account-id",
+		"series":       "16",
+		"account-id":   "account-id",
+		"name":         "my-snap-ctl3",
+		"sequence":     "1",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "my-snap",
+				"id":       "mysnapididididididididididididid",
+				"presence": "required",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	valsets := snapasserts.NewValidationSets()
+
+	// no validation sets
+	vsKeys, _, err := valsets.CheckPresenceRequired(naming.Snap("my-snap"))
+	c.Assert(err, IsNil)
+	c.Check(vsKeys, HasLen, 0)
+
+	c.Assert(valsets.Add(valset1), IsNil)
+	c.Assert(valsets.Add(valset2), IsNil)
+	c.Assert(valsets.Add(valset3), IsNil)
+
+	// sanity
+	c.Assert(valsets.Conflict(), IsNil)
+
+	vsKeys, rev, err := valsets.CheckPresenceRequired(naming.Snap("my-snap"))
+	c.Assert(err, IsNil)
+	c.Check(rev, DeepEquals, snap.Revision{N: 7})
+	c.Check(vsKeys, DeepEquals, []string{"16/account-id/my-snap-ctl/1", "16/account-id/my-snap-ctl2/2", "16/account-id/my-snap-ctl3/1"})
+
+	vsKeys, rev, err = valsets.CheckPresenceRequired(naming.NewSnapRef("my-snap", "mysnapididididididididididididid"))
+	c.Assert(err, IsNil)
+	c.Check(rev, DeepEquals, snap.Revision{N: 7})
+	c.Check(vsKeys, DeepEquals, []string{"16/account-id/my-snap-ctl/1", "16/account-id/my-snap-ctl2/2", "16/account-id/my-snap-ctl3/1"})
+
+	// other-snap is not required
+	vsKeys, rev, err = valsets.CheckPresenceRequired(naming.Snap("other-snap"))
+	c.Assert(err, ErrorMatches, `unexpected presence "invalid" for snap "other-snap"`)
+	pr, ok := err.(*snapasserts.PresenceConstraintError)
+	c.Assert(ok, Equals, true)
+	c.Check(pr.SnapName, Equals, "other-snap")
+	c.Check(pr.Presence, Equals, asserts.PresenceInvalid)
+	c.Check(rev, DeepEquals, snap.Revision{N: 0})
+	c.Check(vsKeys, HasLen, 0)
+
+	// unknown snap is not required
+	vsKeys, rev, err = valsets.CheckPresenceRequired(naming.NewSnapRef("unknown-snap", "00000000idididididididididididid"))
+	c.Assert(err, IsNil)
+	c.Check(rev, DeepEquals, snap.Revision{N: 0})
+	c.Check(vsKeys, HasLen, 0)
+
+	// just one set, required but no revision specified
+	valsets = snapasserts.NewValidationSets()
+	c.Assert(valsets.Add(valset3), IsNil)
+	vsKeys, rev, err = valsets.CheckPresenceRequired(naming.Snap("my-snap"))
+	c.Assert(err, IsNil)
+	c.Check(rev, DeepEquals, snap.Revision{N: 0})
+	c.Check(vsKeys, DeepEquals, []string{"16/account-id/my-snap-ctl3/1"})
+}
+
+func (s *validationSetsSuite) TestIsPresenceInvalid(c *C) {
+	valset1 := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "account-id",
+		"series":       "16",
+		"account-id":   "account-id",
+		"name":         "my-snap-ctl",
+		"sequence":     "1",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "my-snap",
+				"id":       "mysnapididididididididididididid",
+				"presence": "invalid",
+			},
+			map[string]interface{}{
+				"name":     "other-snap",
+				"id":       "123456ididididididididididididid",
+				"presence": "optional",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	valset2 := assertstest.FakeAssertion(map[string]interface{}{
+		"type":         "validation-set",
+		"authority-id": "account-id",
+		"series":       "16",
+		"account-id":   "account-id",
+		"name":         "my-snap-ctl2",
+		"sequence":     "2",
+		"snaps": []interface{}{
+			map[string]interface{}{
+				"name":     "my-snap",
+				"id":       "mysnapididididididididididididid",
+				"presence": "invalid",
+			},
+		},
+	}).(*asserts.ValidationSet)
+
+	valsets := snapasserts.NewValidationSets()
+
+	// no validation sets
+	vsKeys, err := valsets.CheckPresenceInvalid(naming.Snap("my-snap"))
+	c.Assert(err, IsNil)
+	c.Check(vsKeys, HasLen, 0)
+
+	c.Assert(valsets.Add(valset1), IsNil)
+	c.Assert(valsets.Add(valset2), IsNil)
+
+	// sanity
+	c.Assert(valsets.Conflict(), IsNil)
+
+	// invalid in two sets
+	vsKeys, err = valsets.CheckPresenceInvalid(naming.Snap("my-snap"))
+	c.Assert(err, IsNil)
+	c.Check(vsKeys, DeepEquals, []string{"16/account-id/my-snap-ctl/1", "16/account-id/my-snap-ctl2/2"})
+
+	vsKeys, err = valsets.CheckPresenceInvalid(naming.NewSnapRef("my-snap", "mysnapididididididididididididid"))
+	c.Assert(err, IsNil)
+	c.Check(vsKeys, DeepEquals, []string{"16/account-id/my-snap-ctl/1", "16/account-id/my-snap-ctl2/2"})
+
+	// other-snap isn't invalid
+	vsKeys, err = valsets.CheckPresenceInvalid(naming.Snap("other-snap"))
+	c.Assert(err, ErrorMatches, `unexpected presence "optional" for snap "other-snap"`)
+	pr, ok := err.(*snapasserts.PresenceConstraintError)
+	c.Assert(ok, Equals, true)
+	c.Check(pr.SnapName, Equals, "other-snap")
+	c.Check(pr.Presence, Equals, asserts.PresenceOptional)
+	c.Check(vsKeys, HasLen, 0)
+
+	vsKeys, err = valsets.CheckPresenceInvalid(naming.NewSnapRef("other-snap", "123456ididididididididididididid"))
+	c.Assert(err, ErrorMatches, `unexpected presence "optional" for snap "other-snap"`)
+	c.Check(vsKeys, HasLen, 0)
+
+	// unknown snap isn't invalid
+	vsKeys, err = valsets.CheckPresenceInvalid(naming.NewSnapRef("unknown-snap", "00000000idididididididididididid"))
+	c.Assert(err, IsNil)
+	c.Check(vsKeys, HasLen, 0)
+}
