@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2020 Canonical Ltd
+ * Copyright (C) 2015-2021 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -23,13 +23,13 @@ import (
 	"bytes"
 	"context"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"net"
 	"net/http"
 	"os"
 	"os/exec"
 	"os/signal"
-	"strconv"
 	"strings"
 	"sync"
 	"time"
@@ -37,7 +37,6 @@ import (
 	"github.com/gorilla/mux"
 	"gopkg.in/tomb.v2"
 
-	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/logger"
@@ -45,9 +44,9 @@ import (
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/overlord"
 	"github.com/snapcore/snapd/overlord/auth"
+	"github.com/snapcore/snapd/overlord/restart"
 	"github.com/snapcore/snapd/overlord/standby"
 	"github.com/snapcore/snapd/overlord/state"
-	"github.com/snapcore/snapd/polkit"
 	"github.com/snapcore/snapd/snapdenv"
 	"github.com/snapcore/snapd/store"
 	"github.com/snapcore/snapd/systemd"
@@ -58,9 +57,11 @@ var ErrRestartSocket = fmt.Errorf("daemon stop requested to wait for socket acti
 var systemdSdNotify = systemd.SdNotify
 
 const (
-	daemonRestartMsg = "system is restarting"
-	systemRestartMsg = "daemon is restarting"
-	socketRestartMsg = "daemon is stopping to wait for socket activation"
+	daemonRestartMsg  = "daemon is restarting"
+	systemRestartMsg  = "system is restarting"
+	systemHaltMsg     = "system is halting"
+	systemPoweroffMsg = "system is powering off"
+	socketRestartMsg  = "daemon is stopping to wait for socket activation"
 )
 
 // A Daemon listens for requests and routes them to the right command
@@ -77,7 +78,7 @@ type Daemon struct {
 	standbyOpinions *standby.StandbyOpinions
 
 	// set to what kind of restart was requested if any
-	requestedRestart state.RestartType
+	requestedRestart restart.RestartType
 	// set to remember that we need to exit the daemon in a way that
 	// prevents systemd from restarting it
 	restartSocket bool
@@ -100,130 +101,19 @@ type Command struct {
 	GET  ResponseFunc
 	PUT  ResponseFunc
 	POST ResponseFunc
-	// can guest GET?
-	GuestOK bool
-	// can non-admin GET?
-	UserOK bool
-	// is this path accessible on the snapd-snap socket?
-	SnapOK bool
-	// this path is only accessible to root
-	RootOnly bool
 
-	// can polkit grant access? set to polkit action ID if so
-	PolkitOK string
+	// Access control.
+	ReadAccess  accessChecker
+	WriteAccess accessChecker
 
 	d *Daemon
 }
 
-type accessResult int
-
-const (
-	accessOK accessResult = iota
-	accessUnauthorized
-	accessForbidden
-	accessCancelled
-)
-
-var polkitCheckAuthorization = polkit.CheckAuthorization
-
-// canAccess checks the following properties:
-//
-// - if the user is `root` everything is allowed
-// - if a user is logged in (via `snap login`) and the command doesn't have RootOnly, everything is allowed
-// - POST/PUT all require `root`, or just `snap login` if not RootOnly
-//
-// Otherwise for GET requests the following parameters are honored:
-// - GuestOK: anyone can access GET
-// - UserOK: any uid on the local system can access GET
-// - RootOnly: only root can access this
-// - SnapOK: a snap can access this via `snapctl`
-func (c *Command) canAccess(r *http.Request, user *auth.UserState) accessResult {
-	if c.RootOnly && (c.UserOK || c.GuestOK || c.SnapOK) {
-		// programming error
-		logger.Panicf("Command can't have RootOnly together with any *OK flag")
-	}
-
-	if user != nil && !c.RootOnly {
-		// Authenticated users do anything not requiring explicit root.
-		return accessOK
-	}
-
-	// isUser means we have a UID for the request
-	isUser := false
-	pid, uid, socket, err := ucrednetGet(r.RemoteAddr)
-	if err == nil {
-		isUser = true
-	} else if err != errNoID {
-		logger.Noticef("unexpected error when attempting to get UID: %s", err)
-		return accessForbidden
-	}
-	isSnap := (socket == dirs.SnapSocket)
-
-	// ensure that snaps can only access SnapOK things
-	if isSnap {
-		if c.SnapOK {
-			return accessOK
-		}
-		return accessUnauthorized
-	}
-
-	// the !RootOnly check is redundant, but belt-and-suspenders
-	if r.Method == "GET" && !c.RootOnly {
-		// Guest and user access restricted to GET requests
-		if c.GuestOK {
-			return accessOK
-		}
-
-		if isUser && c.UserOK {
-			return accessOK
-		}
-	}
-
-	// Remaining admin checks rely on identifying peer uid
-	if !isUser {
-		return accessUnauthorized
-	}
-
-	if uid == 0 {
-		// Superuser does anything.
-		return accessOK
-	}
-
-	if c.RootOnly {
-		return accessUnauthorized
-	}
-
-	if c.PolkitOK != "" {
-		var flags polkit.CheckFlags
-		allowHeader := r.Header.Get(client.AllowInteractionHeader)
-		if allowHeader != "" {
-			if allow, err := strconv.ParseBool(allowHeader); err != nil {
-				logger.Noticef("error parsing %s header: %s", client.AllowInteractionHeader, err)
-			} else if allow {
-				flags |= polkit.CheckAllowInteraction
-			}
-		}
-		// Pass both pid and uid from the peer ucred to avoid pid race
-		if authorized, err := polkitCheckAuthorization(pid, uid, c.PolkitOK, nil, flags); err == nil {
-			if authorized {
-				// polkit says user is authorised
-				return accessOK
-			}
-		} else if err == polkit.ErrDismissed {
-			return accessCancelled
-		} else {
-			logger.Noticef("polkit error: %s", err)
-		}
-	}
-
-	return accessUnauthorized
-}
-
 func (c *Command) ServeHTTP(w http.ResponseWriter, r *http.Request) {
 	st := c.d.state
 	st.Lock()
 	// TODO Look at the error and fail if there's an attempt to authenticate with invalid data.
-	user, _ := UserFromRequest(st, r)
+	user, _ := userFromRequest(st, r)
 	st.Unlock()
 
 	// check if we are in degradedMode
@@ -232,17 +122,10 @@ func (c *Command) ServeHTTP(w http.ResponseWriter, r *http.Request) {
 		return
 	}
 
-	switch c.canAccess(r, user) {
-	case accessOK:
-		// nothing
-	case accessUnauthorized:
-		Unauthorized("access denied").ServeHTTP(w, r)
-		return
-	case accessForbidden:
-		Forbidden("forbidden").ServeHTTP(w, r)
-		return
-	case accessCancelled:
-		AuthCancelled("cancelled").ServeHTTP(w, r)
+	ucred, err := ucrednetGet(r.RemoteAddr)
+	if err != nil && err != errNoID {
+		logger.Noticef("unexpected error when attempting to get UID: %s", err)
+		InternalError(err.Error()).ServeHTTP(w, r)
 		return
 	}
 
@@ -250,33 +133,49 @@ func (c *Command) ServeHTTP(w http.ResponseWriter, r *http.Request) {
 	r = r.WithContext(ctx)
 
 	var rspf ResponseFunc
-	var rsp = MethodNotAllowed("method %q not allowed", r.Method)
+	var access accessChecker
 
 	switch r.Method {
 	case "GET":
 		rspf = c.GET
+		access = c.ReadAccess
 	case "PUT":
 		rspf = c.PUT
+		access = c.WriteAccess
 	case "POST":
 		rspf = c.POST
+		access = c.WriteAccess
 	}
 
-	if rspf != nil {
-		rsp = rspf(c, r, user)
+	if rspf == nil {
+		MethodNotAllowed("method %q not allowed", r.Method).ServeHTTP(w, r)
+		return
 	}
 
-	if rsp, ok := rsp.(*resp); ok {
-		_, rst := st.Restarting()
-		if rst != state.RestartUnset {
-			rsp.Maintenance = maintenanceForRestartType(rst)
-		}
+	if rspe := access.CheckAccess(c.d, r, ucred, user); rspe != nil {
+		rspe.ServeHTTP(w, r)
+		return
+	}
 
-		if rsp.Type != ResponseTypeError {
+	rsp := rspf(c, r, user)
+
+	if srsp, ok := rsp.(StructuredResponse); ok {
+		rjson := srsp.JSON()
+
+		st.Lock()
+		_, rst := restart.Pending(st)
+		st.Unlock()
+		rjson.addMaintenanceFromRestartType(rst)
+
+		if rjson.Type != ResponseTypeError {
 			st.Lock()
 			count, stamp := st.WarningsSummary()
 			st.Unlock()
-			rsp.addWarningsToMeta(count, stamp)
+			rjson.addWarningCount(count, stamp)
 		}
+
+		// serve the updated serialisation
+		rsp = rjson
 	}
 
 	rsp.ServeHTTP(w, r)
@@ -311,7 +210,7 @@ func logit(handler http.Handler) http.Handler {
 		ww := &wrappedWriter{w: w}
 		t0 := time.Now()
 		handler.ServeHTTP(ww, r)
-		t := time.Now().Sub(t0)
+		t := time.Since(t0)
 		url := r.URL.String()
 		if !strings.Contains(url, "/changes/") {
 			logger.Debugf("%s %s %s %s %d", r.RemoteAddr, r.Method, r.URL, t, ww.s)
@@ -456,8 +355,8 @@ func (d *Daemon) Start() error {
 
 	// before serving actual connections remove the maintenance.json file as we
 	// are no longer down for maintenance, this state most closely corresponds
-	// to state.RestartUnset
-	if err := d.updateMaintenanceFile(state.RestartUnset); err != nil {
+	// to restart.RestartUnset
+	if err := d.updateMaintenanceFile(restart.RestartUnset); err != nil {
 		return err
 	}
 
@@ -488,29 +387,39 @@ func (d *Daemon) Start() error {
 }
 
 // HandleRestart implements overlord.RestartBehavior.
-func (d *Daemon) HandleRestart(t state.RestartType) {
+func (d *Daemon) HandleRestart(t restart.RestartType) {
 	d.mu.Lock()
 	defer d.mu.Unlock()
 
-	// die when asked to restart (systemd should get us back up!) etc
-	switch t {
-	case state.RestartDaemon:
-		// save the restart kind to write out a maintenance.json in a bit
-		d.requestedRestart = t
-	case state.RestartSystem, state.RestartSystemNow:
-		// try to schedule a fallback slow reboot already here
-		// in case we get stuck shutting down
-		if err := reboot(rebootWaitTimeout); err != nil {
+	scheduleFallback := func(a rebootAction) {
+		if err := reboot(a, rebootWaitTimeout); err != nil {
 			logger.Noticef("%s", err)
 		}
+	}
 
+	// die when asked to restart (systemd should get us back up!) etc
+	switch t {
+	case restart.RestartDaemon:
 		// save the restart kind to write out a maintenance.json in a bit
 		d.requestedRestart = t
-	case state.RestartSocket:
+	case restart.RestartSystem, restart.RestartSystemNow:
+		// try to schedule a fallback slow reboot already here
+		// in case we get stuck shutting down
+
+		// save the restart kind to write out a maintenance.json in a bit
+		scheduleFallback(rebootReboot)
+		d.requestedRestart = t
+	case restart.RestartSystemHaltNow:
+		scheduleFallback(rebootHalt)
+		d.requestedRestart = t
+	case restart.RestartSystemPoweroffNow:
+		scheduleFallback(rebootPoweroff)
+		d.requestedRestart = t
+	case restart.RestartSocket:
 		// save the restart kind to write out a maintenance.json in a bit
 		d.requestedRestart = t
 		d.restartSocket = true
-	case state.StopDaemon:
+	case restart.StopDaemon:
 		logger.Noticef("stopping snapd as requested")
 	default:
 		logger.Noticef("internal error: restart handler called with unknown restart type: %v", t)
@@ -526,9 +435,9 @@ var (
 	rebootMaxTentatives    = 3
 )
 
-func (d *Daemon) updateMaintenanceFile(rst state.RestartType) error {
+func (d *Daemon) updateMaintenanceFile(rst restart.RestartType) error {
 	// for unset restart, just remove the maintenance.json file
-	if rst == state.RestartUnset {
+	if rst == restart.RestartUnset {
 		err := os.Remove(dirs.SnapdMaintenanceFile)
 		// only return err if the error was something other than the file not
 		// existing
@@ -553,7 +462,7 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	if d.expectedRebootDidNotHappen {
 		// make the reboot retry immediate
 		immediateReboot := true
-		return d.doReboot(sigCh, immediateReboot, rebootRetryWaitTimeout)
+		return d.doReboot(sigCh, restart.RestartSystem, immediateReboot, rebootRetryWaitTimeout)
 	}
 	if d.overlord == nil {
 		return fmt.Errorf("internal error: no Overlord")
@@ -564,10 +473,18 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	// check the state associated with a potential restart with the lock to
 	// prevent races
 	d.mu.Lock()
-	// needsFullReboot is whether the entire system will be rebooted or not as
-	// a consequence of this restart
-	needsFullReboot := (d.requestedRestart == state.RestartSystemNow || d.requestedRestart == state.RestartSystem)
-	immediateReboot := d.requestedRestart == state.RestartSystemNow
+	// needsFullShutdown is whether the entire system will
+	// shutdown or not as a consequence of this request
+	needsFullShutdown := false
+	switch d.requestedRestart {
+	case restart.RestartSystem, restart.RestartSystemNow, restart.RestartSystemHaltNow, restart.RestartSystemPoweroffNow:
+		needsFullShutdown = true
+	}
+	immediateShutdown := false
+	switch d.requestedRestart {
+	case restart.RestartSystemNow, restart.RestartSystemHaltNow, restart.RestartSystemPoweroffNow:
+		immediateShutdown = true
+	}
 	restartSocket := d.restartSocket
 	d.mu.Unlock()
 
@@ -585,7 +502,10 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 		// stop running hooks first
 		// and do it more gracefully if we are restarting
 		hookMgr := d.overlord.HookManager()
-		if ok, _ := d.state.Restarting(); ok {
+		d.state.Lock()
+		ok, _ := restart.Pending(d.state)
+		d.state.Unlock()
+		if ok {
 			logger.Noticef("gracefully waiting for running hooks")
 			hookMgr.GracefullyWaitRunningHooks()
 			logger.Noticef("done waiting for running hooks")
@@ -594,7 +514,7 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 		d.snapListener.Close()
 	}
 
-	if needsFullReboot {
+	if needsFullShutdown {
 		// give time to polling clients to notice restart
 		time.Sleep(rebootNoticeWait)
 	}
@@ -606,7 +526,7 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	d.tomb.Kill(d.serve.Shutdown(ctx))
 	cancel()
 
-	if !needsFullReboot {
+	if !needsFullShutdown {
 		// tell systemd that we are stopping
 		systemdSdNotify("STOPPING=1")
 	}
@@ -637,7 +557,7 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 			// because we already scheduled a slow shutdown and
 			// exiting here will just restart snapd (via systemd)
 			// which will lead to confusing results.
-			if needsFullReboot {
+			if needsFullShutdown {
 				logger.Noticef("WARNING: cannot stop daemon: %v", err)
 			} else {
 				return err
@@ -645,8 +565,8 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 		}
 	}
 
-	if needsFullReboot {
-		return d.doReboot(sigCh, immediateReboot, rebootWaitTimeout)
+	if needsFullShutdown {
+		return d.doReboot(sigCh, d.requestedRestart, immediateShutdown, rebootWaitTimeout)
 	}
 
 	if d.restartSocket {
@@ -686,18 +606,25 @@ func (d *Daemon) rebootDelay(immediate bool) (time.Duration, error) {
 	return rebootDelay, nil
 }
 
-func (d *Daemon) doReboot(sigCh chan<- os.Signal, immediate bool, waitTimeout time.Duration) error {
+func (d *Daemon) doReboot(sigCh chan<- os.Signal, rst restart.RestartType, immediate bool, waitTimeout time.Duration) error {
 	rebootDelay, err := d.rebootDelay(immediate)
 	if err != nil {
 		return err
 	}
+	action := rebootReboot
+	switch rst {
+	case restart.RestartSystemHaltNow:
+		action = rebootHalt
+	case restart.RestartSystemPoweroffNow:
+		action = rebootPoweroff
+	}
 	// ask for shutdown and wait for it to happen.
 	// if we exit snapd will be restared by systemd
-	if err := reboot(rebootDelay); err != nil {
+	if err := reboot(action, rebootDelay); err != nil {
 		return err
 	}
 	// wait for reboot to happen
-	logger.Noticef("Waiting for system reboot")
+	logger.Noticef("Waiting for %s", action)
 	if sigCh != nil {
 		signal.Stop(sigCh)
 		if len(sigCh) > 0 {
@@ -707,17 +634,56 @@ func (d *Daemon) doReboot(sigCh chan<- os.Signal, immediate bool, waitTimeout ti
 		close(sigCh)
 	}
 	time.Sleep(waitTimeout)
-	return fmt.Errorf("expected reboot did not happen")
+	return fmt.Errorf("expected %s did not happen", action)
 }
 
-var shutdownMsg = i18n.G("reboot scheduled to update the system")
+var (
+	shutdownMsg = i18n.G("reboot scheduled to update the system")
+	haltMsg     = i18n.G("system halt scheduled")
+	poweroffMsg = i18n.G("system poweroff scheduled")
+)
 
-func rebootImpl(rebootDelay time.Duration) error {
+type rebootAction int
+
+func (a rebootAction) String() string {
+	switch a {
+	case rebootReboot:
+		return "system reboot"
+	case rebootHalt:
+		return "system halt"
+	case rebootPoweroff:
+		return "system poweroff"
+	default:
+		panic(fmt.Sprintf("unknown reboot action %d", a))
+	}
+}
+
+const (
+	rebootReboot rebootAction = iota
+	rebootHalt
+	rebootPoweroff
+)
+
+func rebootImpl(action rebootAction, rebootDelay time.Duration) error {
 	if rebootDelay < 0 {
 		rebootDelay = 0
 	}
 	mins := int64(rebootDelay / time.Minute)
-	cmd := exec.Command("shutdown", "-r", fmt.Sprintf("+%d", mins), shutdownMsg)
+	var arg, msg string
+	switch action {
+	case rebootReboot:
+		arg = "-r"
+		msg = shutdownMsg
+	case rebootHalt:
+		arg = "--halt"
+		msg = haltMsg
+	case rebootPoweroff:
+		arg = "--poweroff"
+		msg = poweroffMsg
+	default:
+		return fmt.Errorf("unknown reboot action: %v", action)
+	}
+	cmd := exec.Command("shutdown", arg, fmt.Sprintf("+%d", mins), msg)
 	if out, err := cmd.CombinedOutput(); err != nil {
 		return osutil.OutputErr(out, err)
 	}
@@ -742,6 +708,8 @@ func (d *Daemon) RebootAsExpected(st *state.State) error {
 	return nil
 }
 
+var errExpectedReboot = errors.New("expected reboot did not happen")
+
 // RebootDidNotHappen implements part of overlord.RestartBehavior.
 func (d *Daemon) RebootDidNotHappen(st *state.State) error {
 	var nTentative int
@@ -753,7 +721,7 @@ func (d *Daemon) RebootDidNotHappen(st *state.State) error {
 	if nTentative > rebootMaxTentatives {
 		// giving up, proceed normally, some in-progress refresh
 		// might get rolled back!!
-		st.ClearReboot()
+		restart.ClearReboot(st)
 		clearReboot(st)
 		logger.Noticef("snapd was restarted while a system restart was expected, snapd retried to schedule and waited again for a system restart %d times and is giving up", rebootMaxTentatives)
 		return nil
@@ -761,14 +729,14 @@ func (d *Daemon) RebootDidNotHappen(st *state.State) error {
 	st.Set("daemon-system-restart-tentative", nTentative)
 	d.state = st
 	logger.Noticef("snapd was restarted while a system restart was expected, snapd will try to schedule and wait for a system restart again (tenative %d/%d)", nTentative, rebootMaxTentatives)
-	return state.ErrExpectedReboot
+	return errExpectedReboot
 }
 
 // New Daemon
 func New() (*Daemon, error) {
 	d := &Daemon{}
 	ovld, err := overlord.New(d)
-	if err == state.ErrExpectedReboot {
+	if err == errExpectedReboot {
 		// we proceed without overlord until we reach Stop
 		// where we will schedule and wait again for a system restart.
 		// ATM we cannot do that in New because we need to satisfy
