@@ -140,16 +140,20 @@ func (s *baseDeclSuite) TestAutoConnection(c *C) {
 	// these have more complex or in flux policies and have their
 	// own separate tests
 	snowflakes := map[string]bool{
-		"content":       true,
-		"core-support":  true,
-		"home":          true,
-		"lxd-support":   true,
-		"snapd-control": true,
-		"dummy":         true,
+		"content":            true,
+		"core-support":       true,
+		"home":               true,
+		"lxd-support":        true,
+		"microstack-support": true,
+		"multipass-support":  true,
+		"packagekit-control": true,
+		"snapd-control":      true,
+		"dummy":              true,
 	}
 
 	// these simply auto-connect, anything else doesn't
 	autoconnect := map[string]bool{
+		"audio-playback":          true,
 		"browser-support":         true,
 		"desktop":                 true,
 		"desktop-legacy":          true,
@@ -162,7 +166,6 @@ func (s *baseDeclSuite) TestAutoConnection(c *C) {
 		"online-accounts-service": true,
 		"opengl":                  true,
 		"optical-drive":           true,
-		"pulseaudio":              true,
 		"screen-inhibit-control":  true,
 		"ubuntu-download-manager": true,
 		"unity7":                  true,
@@ -181,9 +184,10 @@ func (s *baseDeclSuite) TestAutoConnection(c *C) {
 
 		// check base declaration
 		cand := s.connectCand(c, iface.Name(), "", "")
-		err := cand.CheckAutoConnect()
+		arity, err := cand.CheckAutoConnect()
 		if expected {
 			c.Check(err, IsNil, comm)
+			c.Check(arity.SlotsPerPlugAny(), Equals, false)
 		} else {
 			c.Check(err, NotNil, comm)
 		}
@@ -200,6 +204,8 @@ func (s *baseDeclSuite) TestAutoConnectPlugSlot(c *C) {
 		"content":         true,
 		"home":            true,
 		"lxd-support":     true,
+		// netlink-driver needs the family-name attributes to match
+		"netlink-driver": true,
 	}
 
 	for _, iface := range all {
@@ -214,11 +220,12 @@ func (s *baseDeclSuite) TestInterimAutoConnectionHome(c *C) {
 	restore := release.MockOnClassic(true)
 	defer restore()
 	cand := s.connectCand(c, "home", "", "")
-	err := cand.CheckAutoConnect()
+	arity, err := cand.CheckAutoConnect()
 	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
 
 	release.OnClassic = false
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, ErrorMatches, `auto-connection denied by slot rule of interface \"home\"`)
 }
 
@@ -235,14 +242,14 @@ plugs:
 	err := cand.Check()
 	c.Check(err, NotNil)
 
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 
 	release.OnClassic = false
 	err = cand.Check()
 	c.Check(err, NotNil)
 
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 }
 
@@ -259,21 +266,22 @@ plugs:
 	c.Check(err, IsNil)
 
 	// Same as TestInterimAutoConnectionHome()
-	err = cand.CheckAutoConnect()
+	arity, err := cand.CheckAutoConnect()
 	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
 
 	release.OnClassic = false
 	err = cand.Check()
 	c.Check(err, IsNil)
 
 	// Same as TestInterimAutoConnectionHome()
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionSnapdControl(c *C) {
 	cand := s.connectCand(c, "snapd-control", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"snapd-control\"")
 
@@ -285,15 +293,16 @@ plugs:
 
 	lxdDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = lxdDecl
-	err = cand.CheckAutoConnect()
+	arity, err := cand.CheckAutoConnect()
 	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionContent(c *C) {
 	// random snaps cannot connect with content
 	// (Sanitize* will now also block this)
 	cand := s.connectCand(c, "content", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 
 	slotDecl1 := s.mockSnapDecl(c, "slot-snap", "slot-snap-id", "pub1", "")
@@ -318,13 +327,14 @@ plugs:
 `)
 	cand.SlotSnapDeclaration = slotDecl1
 	cand.PlugSnapDeclaration = plugDecl1
-	err = cand.CheckAutoConnect()
+	arity, err := cand.CheckAutoConnect()
 	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
 
 	// different publisher, same content
 	cand.SlotSnapDeclaration = slotDecl1
 	cand.PlugSnapDeclaration = plugDecl2
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 
 	// same publisher, different content
@@ -344,14 +354,14 @@ plugs:
 `)
 	cand.SlotSnapDeclaration = slotDecl1
 	cand.PlugSnapDeclaration = plugDecl1
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionLxdSupportOverride(c *C) {
 	// by default, don't auto-connect
 	cand := s.connectCand(c, "lxd-support", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 
 	plugsSlots := `
@@ -362,7 +372,7 @@ plugs:
 
 	lxdDecl := s.mockSnapDecl(c, "lxd", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = lxdDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
@@ -376,14 +386,14 @@ plugs:
 
 	lxdDecl := s.mockSnapDecl(c, "notlxd", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = lxdDecl
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection not allowed by plug rule of interface \"lxd-support\" for \"notlxd\" snap")
 }
 
 func (s *baseDeclSuite) TestAutoConnectionKernelModuleControlOverride(c *C) {
 	cand := s.connectCand(c, "kernel-module-control", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"kernel-module-control\"")
 
@@ -395,13 +405,13 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionDockerSupportOverride(c *C) {
 	cand := s.connectCand(c, "docker-support", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"docker-support\"")
 
@@ -413,13 +423,13 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionClassicSupportOverride(c *C) {
 	cand := s.connectCand(c, "classic-support", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"classic-support\"")
 
@@ -431,13 +441,13 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "classic", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionKubernetesSupportOverride(c *C) {
 	cand := s.connectCand(c, "kubernetes-support", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"kubernetes-support\"")
 
@@ -449,13 +459,30 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
+func (s *baseDeclSuite) TestAutoConnectionMicroStackSupportOverride(c *C) {
+	cand := s.connectCand(c, "microstack-support", "", "")
+	_, err := cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"microstack-support\"")
+
+	plugsSlots := `
+plugs:
+  microstack-support:
+    allow-auto-connection: true
+`
+
+	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
+	cand.PlugSnapDeclaration = snapDecl
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+}
 func (s *baseDeclSuite) TestAutoConnectionGreengrassSupportOverride(c *C) {
 	cand := s.connectCand(c, "greengrass-support", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"greengrass-support\"")
 
@@ -467,13 +494,31 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+}
+
+func (s *baseDeclSuite) TestAutoConnectionMultipassSupportOverride(c *C) {
+	cand := s.connectCand(c, "multipass-support", "", "")
+	_, err := cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"multipass-support\"")
+
+	plugsSlots := `
+plugs:
+  multipass-support:
+    allow-auto-connection: true
+`
+
+	snapDecl := s.mockSnapDecl(c, "multipass-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
+	cand.PlugSnapDeclaration = snapDecl
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
 func (s *baseDeclSuite) TestAutoConnectionBlockDevicesOverride(c *C) {
 	cand := s.connectCand(c, "block-devices", "", "")
-	err := cand.CheckAutoConnect()
+	_, err := cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"block-devices\"")
 
@@ -485,7 +530,25 @@ plugs:
 
 	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
 	cand.PlugSnapDeclaration = snapDecl
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+}
+
+func (s *baseDeclSuite) TestAutoConnectionPackagekitControlOverride(c *C) {
+	cand := s.connectCand(c, "packagekit-control", "", "")
+	_, err := cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"packagekit-control\"")
+
+	plugsSlots := `
+plugs:
+  packagekit-control:
+    allow-auto-connection: true
+`
+
+	snapDecl := s.mockSnapDecl(c, "some-snap", "J60k4JY0HppjwOjW8dZdYc8obXKxujRu", "canonical", plugsSlots)
+	cand.PlugSnapDeclaration = snapDecl
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, IsNil)
 }
 
@@ -522,8 +585,9 @@ plugs:
 
 		cand := s.connectCand(c, iface.Name(), "", "")
 		cand.PlugSnapDeclaration = snapDecl
-		err := cand.CheckAutoConnect()
+		arity, err := cand.CheckAutoConnect()
 		c.Check(err, IsNil)
+		c.Check(arity.SlotsPerPlugAny(), Equals, false)
 	}
 }
 
@@ -536,50 +600,67 @@ plugs:
 // ATM a nil entry means even stricter rules that would need be tested
 // separately and whose implementation is in flux for now
 var (
-	unconstrained = []string{"core", "kernel", "gadget", "app"}
-
 	slotInstallation = map[string][]string{
 		// other
-		"adb-support":             {"core"},
-		"autopilot-introspection": {"core"},
-		"avahi-control":           {"app", "core"},
-		"avahi-observe":           {"app", "core"},
-		"bluez":                   {"app", "core"},
-		"bool-file":               {"core", "gadget"},
-		"browser-support":         {"core"},
-		"content":                 {"app", "gadget"},
-		"core-support":            {"core"},
-		"dbus":                    {"app"},
-		"docker-support":          {"core"},
-		"fwupd":                   {"app"},
-		"gpio":                    {"core", "gadget"},
-		"greengrass-support":      {"core"},
-		"hidraw":                  {"core", "gadget"},
-		"i2c":                     {"core", "gadget"},
-		"iio":                     {"core", "gadget"},
-		"kubernetes-support":      {"core"},
-		"location-control":        {"app"},
-		"location-observe":        {"app"},
-		"lxd-support":             {"core"},
-		"maliit":                  {"app"},
-		"media-hub":               {"app", "core"},
-		"mir":                     {"app"},
-		"modem-manager":           {"app", "core"},
-		"mpris":                   {"app"},
-		"network-manager":         {"app", "core"},
-		"network-status":          {"app"},
-		"ofono":                   {"app", "core"},
-		"online-accounts-service": {"app"},
-		"ppp":         {"core"},
-		"pulseaudio":  {"app", "core"},
-		"serial-port": {"core", "gadget"},
-		"spi":         {"core", "gadget"},
-		"storage-framework-service": {"app"},
+		"adb-support":               {"core"},
+		"audio-playback":            {"app", "core"},
+		"audio-record":              {"app", "core"},
+		"autopilot-introspection":   {"core"},
+		"avahi-control":             {"app", "core"},
+		"avahi-observe":             {"app", "core"},
+		"bluez":                     {"app", "core"},
+		"bool-file":                 {"core", "gadget"},
+		"browser-support":           {"core"},
+		"content":                   {"app", "gadget"},
+		"core-support":              {"core"},
+		"cups":                      {"app"},
+		"cups-control":              {"app", "core"},
+		"dbus":                      {"app"},
+		"docker-support":            {"core"},
+		"desktop-launch":            {"core"},
+		"dsp":                       {"core", "gadget"},
 		"dummy":                     {"app"},
+		"fwupd":                     {"app", "core"},
+		"gpio":                      {"core", "gadget"},
+		"gpio-control":              {"core"},
+		"greengrass-support":        {"core"},
+		"hidraw":                    {"core", "gadget"},
+		"i2c":                       {"core", "gadget"},
+		"iio":                       {"core", "gadget"},
+		"kernel-module-load":        {"core"},
+		"kubernetes-support":        {"core"},
+		"location-control":          {"app"},
+		"location-observe":          {"app"},
+		"lxd-support":               {"core"},
+		"maliit":                    {"app"},
+		"media-hub":                 {"app", "core"},
+		"mir":                       {"app"},
+		"microstack-support":        {"core"},
+		"modem-manager":             {"app", "core"},
+		"mount-control":             {"core"},
+		"mpris":                     {"app"},
+		"netlink-driver":            {"core", "gadget"},
+		"network-manager":           {"app", "core"},
+		"network-manager-observe":   {"app", "core"},
+		"network-status":            {"core"},
+		"ofono":                     {"app", "core"},
+		"online-accounts-service":   {"app"},
+		"power-control":             {"core"},
+		"ppp":                       {"core"},
+		"pulseaudio":                {"app", "core"},
+		"pwm":                       {"core", "gadget"},
+		"qualcomm-ipc-router":       {"core"},
+		"raw-volume":                {"core", "gadget"},
+		"scsi-generic":              {"core"},
+		"sd-control":                {"core"},
+		"serial-port":               {"core", "gadget"},
+		"spi":                       {"core", "gadget"},
+		"storage-framework-service": {"app"},
 		"thumbnailer-service":       {"app"},
 		"ubuntu-download-manager":   {"app"},
 		"udisks2":                   {"app", "core"},
 		"uhid":                      {"core"},
+		"uio":                       {"core", "gadget"},
 		"unity8":                    {"app"},
 		"unity8-calendar":           {"app"},
 		"unity8-contacts":           {"app"},
@@ -590,6 +671,7 @@ var (
 		"classic-support": nil,
 		"docker":          nil,
 		"lxd":             nil,
+		"shared-memory":   nil,
 	}
 
 	restrictedPlugInstallation = map[string][]string{
@@ -609,12 +691,10 @@ func (s *baseDeclSuite) TestSlotInstallation(c *C) {
 
 	for _, iface := range all {
 		types, ok := slotInstallation[iface.Name()]
-		compareWithSanitize := false
 		if !ok { // common ones, only core can install them,
-			// their plain BeforePrepareSlot checked for that
 			types = []string{"core"}
-			compareWithSanitize = true
 		}
+
 		if types == nil {
 			// snowflake needs to be tested specially
 			continue
@@ -622,7 +702,6 @@ func (s *baseDeclSuite) TestSlotInstallation(c *C) {
 		for name, snapType := range snapTypeMap {
 			ok := strutil.ListContains(types, name)
 			ic := s.installSlotCand(c, iface.Name(), snapType, ``)
-			slotInfo := ic.Snap.Slots[iface.Name()]
 			err := ic.Check()
 			comm := Commentf("%s by %s snap", iface.Name(), name)
 			if ok {
@@ -630,14 +709,6 @@ func (s *baseDeclSuite) TestSlotInstallation(c *C) {
 			} else {
 				c.Check(err, NotNil, comm)
 			}
-			if compareWithSanitize {
-				sanitizeErr := interfaces.BeforePrepareSlot(iface, slotInfo)
-				if err == nil {
-					c.Check(sanitizeErr, IsNil, comm)
-				} else {
-					c.Check(sanitizeErr, NotNil, comm)
-				}
-			}
 		}
 	}
 
@@ -660,14 +731,28 @@ func (s *baseDeclSuite) TestPlugInstallation(c *C) {
 	restricted := map[string]bool{
 		"block-devices":         true,
 		"classic-support":       true,
+		"desktop-launch":        true,
+		"dm-crypt":              true,
 		"docker-support":        true,
 		"greengrass-support":    true,
+		"gpio-control":          true,
+		"ion-memory-control":    true,
 		"kernel-module-control": true,
+		"kernel-module-load":    true,
 		"kubernetes-support":    true,
 		"lxd-support":           true,
+		"microstack-support":    true,
+		"mount-control":         true,
+		"multipass-support":     true,
+		"packagekit-control":    true,
 		"personal-files":        true,
+		"sd-control":            true,
+		"snap-refresh-control":  true,
+		"snap-themes-control":   true,
 		"snapd-control":         true,
 		"system-files":          true,
+		"tee":                   true,
+		"uinput":                true,
 		"unity8":                true,
 	}
 
@@ -708,6 +793,7 @@ func (s *baseDeclSuite) TestConnection(c *C) {
 	// case-by-case basis
 	noconnect := map[string]bool{
 		"content":                   true,
+		"cups":                      true,
 		"docker":                    true,
 		"fwupd":                     true,
 		"location-control":          true,
@@ -715,8 +801,9 @@ func (s *baseDeclSuite) TestConnection(c *C) {
 		"lxd":                       true,
 		"maliit":                    true,
 		"mir":                       true,
-		"network-status":            true,
 		"online-accounts-service":   true,
+		"raw-volume":                true,
+		"shared-memory":             true,
 		"storage-framework-service": true,
 		"thumbnailer-service":       true,
 		"ubuntu-download-manager":   true,
@@ -749,6 +836,7 @@ func (s *baseDeclSuite) TestConnectionOnClassic(c *C) {
 	// connecting with these interfaces needs to be allowed on
 	// case-by-case basis when not on classic
 	noconnect := map[string]bool{
+		"audio-record":    true,
 		"modem-manager":   true,
 		"network-manager": true,
 		"ofono":           true,
@@ -778,6 +866,103 @@ func (s *baseDeclSuite) TestConnectionOnClassic(c *C) {
 	}
 }
 
+func (s *baseDeclSuite) TestConnectionImplicitOnClassicOrAppSnap(c *C) {
+	restore := release.MockOnClassic(false)
+	defer restore()
+
+	all := builtin.Interfaces()
+
+	// These interfaces represent when the interface might be implicit on
+	// classic or when the interface is provided via an app snap. As such,
+	// they all share the following:
+	//
+	// - implicitOnCore: false
+	// - implicitOnClassis: true
+	// - base declaration uses:
+	//     allow-installation:
+	//       slot-snap-type:
+	//         - app
+	//         - core
+	//     deny-connection:
+	//       on-classic: false
+	//     deny-auto-connection: true|false|unspecified
+	//
+	// connecting with these interfaces needs to be allowed on
+	// case-by-case basis when not on classic
+	ifaces := map[string]bool{
+		"audio-playback":  true,
+		"audio-record":    true,
+		"cups-control":    true,
+		"modem-manager":   true,
+		"network-manager": true,
+		"ofono":           true,
+		"pulseaudio":      true,
+		"upower-observe":  true,
+	}
+
+	for _, iface := range all {
+		if !ifaces[iface.Name()] {
+			continue
+		}
+		comm := Commentf(iface.Name())
+
+		// verify the interface is setup as expected wrt
+		// implicitOnCore, implicitOnClassic, no plugs and has
+		// expected slots (ignoring AutoConnection)
+		si := interfaces.StaticInfoOf(iface)
+		c.Assert(si.ImplicitOnCore, Equals, false)
+		c.Assert(si.ImplicitOnClassic, Equals, true)
+
+		c.Assert(s.baseDecl.PlugRule(iface.Name()), IsNil)
+
+		sr := s.baseDecl.SlotRule(iface.Name())
+		c.Assert(sr, Not(IsNil))
+		c.Assert(sr.AllowInstallation, HasLen, 1)
+		c.Check(sr.AllowInstallation[0].SlotSnapTypes, DeepEquals, []string{"app", "core"}, comm)
+		c.Assert(sr.DenyConnection, HasLen, 1)
+		c.Check(sr.DenyConnection[0].OnClassic, DeepEquals, &asserts.OnClassicConstraint{Classic: false}, comm)
+
+		for _, onClassic := range []bool{true, false} {
+			release.OnClassic = onClassic
+
+			for _, implicit := range []bool{true, false} {
+				// When implicitOnCore is false, there is
+				// nothing to test on Core
+				if implicit && !onClassic {
+					continue
+				}
+
+				snapType := "app"
+				if implicit {
+					snapType = "os"
+				}
+				slotYaml := fmt.Sprintf(`name: slot-snap
+version: 0
+type: %s
+slots:
+  %s:
+`, snapType, iface.Name())
+
+				// XXX: eventually 'onClassic && !implicit' but
+				// the current declaration allows connection on
+				// Core when 'type: app'. See:
+				// https://github.com/snapcore/snapd/pull/8920/files#r471678529
+				expected := onClassic
+
+				// check base declaration
+				cand := s.connectCand(c, iface.Name(), slotYaml, "")
+				err := cand.Check()
+
+				if expected {
+					c.Check(err, IsNil, comm)
+				} else {
+					c.Check(err, NotNil, comm)
+				}
+			}
+		}
+	}
+}
+
 func (s *baseDeclSuite) TestSanity(c *C) {
 	all := builtin.Interfaces()
 
@@ -786,17 +971,33 @@ func (s *baseDeclSuite) TestSanity(c *C) {
 	// listed here to make sure that was a conscious decision
 	bothSides := map[string]bool{
 		"block-devices":         true,
+		"audio-playback":        true,
 		"classic-support":       true,
 		"core-support":          true,
+		"desktop-launch":        true,
+		"dm-crypt":              true,
 		"docker-support":        true,
 		"greengrass-support":    true,
+		"gpio-control":          true,
+		"ion-memory-control":    true,
 		"kernel-module-control": true,
+		"kernel-module-load":    true,
 		"kubernetes-support":    true,
 		"lxd-support":           true,
+		"microstack-support":    true,
+		"mount-control":         true,
+		"multipass-support":     true,
+		"packagekit-control":    true,
 		"personal-files":        true,
+		"sd-control":            true,
+		"shared-memory":         true,
+		"snap-refresh-control":  true,
+		"snap-themes-control":   true,
 		"snapd-control":         true,
 		"system-files":          true,
+		"tee":                   true,
 		"udisks2":               true,
+		"uinput":                true,
 		"unity8":                true,
 		"wayland":               true,
 	}
@@ -906,7 +1107,7 @@ plugs:
 	err := cand.Check()
 	c.Check(err, NotNil)
 
-	err = cand.CheckAutoConnect()
+	_, err = cand.CheckAutoConnect()
 	c.Check(err, NotNil)
 }
 
@@ -953,7 +1154,7 @@ plugs:
 		cand := s.connectCand(c, "optical-drive", "", plugYaml)
 		err := cand.Check()
 		c.Check(err, checker)
-		err = cand.CheckAutoConnect()
+		_, err = cand.CheckAutoConnect()
 		c.Check(err, checker)
 	}
 
@@ -964,3 +1165,107 @@ plugs:
 		checkOpticalDriveAutoConnect(plugYaml, NotNil)
 	}
 }
+
+func (s *baseDeclSuite) TestRawVolumeOverride(c *C) {
+	slotYaml := `name: slot-snap
+type: gadget
+version: 0
+slots:
+  raw-volume:
+    path: /dev/mmcblk0p1
+`
+	slotSnap := snaptest.MockInfo(c, slotYaml, nil)
+	// mock a well-formed slot snap decl with SnapID
+	slotSnapDecl := s.mockSnapDecl(c, "slot-snap", "slotsnapidididididididididididid", "canonical", "")
+
+	plugYaml := `name: plug-snap
+version: 0
+plugs:
+  raw-volume:
+`
+	plugSnap := snaptest.MockInfo(c, plugYaml, nil)
+
+	// no plug-side declaration
+	cand := &policy.ConnectCandidate{
+		Plug:                interfaces.NewConnectedPlug(plugSnap.Plugs["raw-volume"], nil, nil),
+		Slot:                interfaces.NewConnectedSlot(slotSnap.Slots["raw-volume"], nil, nil),
+		SlotSnapDeclaration: slotSnapDecl,
+		BaseDeclaration:     s.baseDecl,
+	}
+
+	err := cand.Check()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "connection denied by slot rule of interface \"raw-volume\"")
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection denied by slot rule of interface \"raw-volume\"")
+
+	// specific plug-side declaration for connection only
+	plugsOverride := `
+plugs:
+  raw-volume:
+    allow-connection:
+      slot-snap-id:
+        - slotsnapidididididididididididid
+    allow-auto-connection: false
+`
+	plugSnapDecl := s.mockSnapDecl(c, "plug-snap", "plugsnapidididididididididididid", "canonical", plugsOverride)
+	cand.PlugSnapDeclaration = plugSnapDecl
+	err = cand.Check()
+	c.Check(err, IsNil)
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection not allowed by plug rule of interface \"raw-volume\" for \"plug-snap\" snap")
+
+	// specific plug-side declaration for connection and auto-connection
+	plugsOverride = `
+plugs:
+  raw-volume:
+    allow-connection:
+      slot-snap-id:
+        - slotsnapidididididididididididid
+    allow-auto-connection:
+      slot-snap-id:
+        - slotsnapidididididididididididid
+`
+	plugSnapDecl = s.mockSnapDecl(c, "plug-snap", "plugsnapidididididididididididid", "canonical", plugsOverride)
+	cand.PlugSnapDeclaration = plugSnapDecl
+	err = cand.Check()
+	c.Check(err, IsNil)
+	arity, err := cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
+
+	// blanket allow for connection and auto-connection to any slotting snap
+	plugsOverride = `
+plugs:
+  raw-volume:
+    allow-connection: true
+    allow-auto-connection: true
+`
+	plugSnapDecl = s.mockSnapDecl(c, "some-snap", "plugsnapidididididididididididid", "canonical", plugsOverride)
+	cand.PlugSnapDeclaration = plugSnapDecl
+	err = cand.Check()
+	c.Check(err, IsNil)
+	arity, err = cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+	c.Check(arity.SlotsPerPlugAny(), Equals, false)
+}
+
+func (s *baseDeclSuite) TestAutoConnectionDesktopLaunchOverride(c *C) {
+	cand := s.connectCand(c, "desktop-launch", "", "")
+	_, err := cand.CheckAutoConnect()
+	c.Check(err, NotNil)
+	c.Assert(err, ErrorMatches, "auto-connection denied by plug rule of interface \"desktop-launch\"")
+
+	plugsSlots := `
+plugs:
+  desktop-launch:
+    allow-auto-connection: true
+`
+
+	snapDecl := s.mockSnapDecl(c, "some-snap", "some-snap-with-desktop-launch-id", "canonical", plugsSlots)
+	cand.PlugSnapDeclaration = snapDecl
+	_, err = cand.CheckAutoConnect()
+	c.Check(err, IsNil)
+}
