@@ -29,6 +29,7 @@ import (
 	"gopkg.in/tomb.v2"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/snapshotstate/backend"
@@ -196,15 +197,10 @@ func prepareSave(task *state.Task) (snapshot *snapshotSetup, cur *snap.Info, cfg
 	snapshot.Filename = filename(snapshot.SetID, cur)
 	task.Set("snapshot-setup", &snapshot)
 
-	rawCfg, err := configGetSnapConfig(st, snapshot.Snap)
+	cfg, err = unmarshalSnapConfig(st, snapshot.Snap)
 	if err != nil {
 		return nil, nil, nil, err
 	}
-	if rawCfg != nil {
-		if err := json.Unmarshal(*rawCfg, &cfg); err != nil {
-			return nil, nil, nil, err
-		}
-	}
 
 	// this should be done last because of it modifies the state and the caller needs to undo this if other operation fails.
 	if snapshot.Auto {
@@ -225,9 +221,17 @@ func doSave(task *state.Task, tomb *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	_, err = backendSave(tomb.Context(nil), snapshot.SetID, cur, cfg, snapshot.Users)
+	st := task.State()
+
+	st.Lock()
+	opts, err := snapstate.GetSnapDirOptions(st)
+	st.Unlock()
+	if err != nil {
+		return err
+	}
+
+	_, err = backendSave(tomb.Context(nil), snapshot.SetID, cur, cfg, snapshot.Users, opts)
 	if err != nil {
-		st := task.State()
 		st.Lock()
 		defer st.Unlock()
 		removeSnapshotState(st, snapshot.SetID)
@@ -247,17 +251,10 @@ func prepareRestore(task *state.Task) (snapshot *snapshotSetup, oldCfg map[strin
 		return nil, nil, nil, taskGetErrMsg(task, err, "snapshot")
 	}
 
-	rawCfg, err := configGetSnapConfig(st, snapshot.Snap)
+	oldCfg, err = unmarshalSnapConfig(st, snapshot.Snap)
 	if err != nil {
-		return nil, nil, nil, fmt.Errorf("internal error: cannot obtain current snap config for snapshot restore: %v", err)
+		return nil, nil, nil, err
 	}
-
-	if rawCfg != nil {
-		if err := json.Unmarshal(*rawCfg, &oldCfg); err != nil {
-			return nil, nil, nil, fmt.Errorf("internal error: cannot decode current snap config: %v", err)
-		}
-	}
-
 	reader, err = backendOpen(snapshot.Filename, backend.ExtractFnameSetID)
 	if err != nil {
 		return nil, nil, nil, fmt.Errorf("cannot open snapshot: %v", err)
@@ -267,6 +264,36 @@ func prepareRestore(task *state.Task) (snapshot *snapshotSetup, oldCfg map[strin
 	return snapshot, oldCfg, reader, nil
 }
 
+// marshalSnapConfig encodes cfg to JSON and returns raw JSON message, unless
+// cfg is nil - in this case nil is returned.
+func marshalSnapConfig(cfg map[string]interface{}) (*json.RawMessage, error) {
+	if cfg == nil {
+		// do not marshal nil - this would result in "null" raw message which
+		// we want to avoid.
+		return nil, nil
+	}
+	buf, err := json.Marshal(cfg)
+	if err != nil {
+		return nil, err
+	}
+	raw := (*json.RawMessage)(&buf)
+	return raw, err
+}
+
+func unmarshalSnapConfig(st *state.State, snapName string) (map[string]interface{}, error) {
+	rawCfg, err := configGetSnapConfig(st, snapName)
+	if err != nil {
+		return nil, fmt.Errorf("internal error: cannot obtain current snap config: %v", err)
+	}
+	var cfg map[string]interface{}
+	if rawCfg != nil {
+		if err := json.Unmarshal(*rawCfg, &cfg); err != nil {
+			return nil, fmt.Errorf("internal error: cannot decode current snap config: %v", err)
+		}
+	}
+	return cfg, nil
+}
+
 func doRestore(task *state.Task, tomb *tomb.Tomb) error {
 	snapshot, oldCfg, reader, err := prepareRestore(task)
 	if err != nil {
@@ -281,12 +308,19 @@ func doRestore(task *state.Task, tomb *tomb.Tomb) error {
 		task.Logf(format, args...)
 	}
 
-	restoreState, err := backendRestore(reader, tomb.Context(nil), snapshot.Current, snapshot.Users, logf)
+	st.Lock()
+	opts, err := snapstate.GetSnapDirOptions(st)
+	st.Unlock()
 	if err != nil {
 		return err
 	}
 
-	buf, err := json.Marshal(reader.Conf)
+	restoreState, err := backendRestore(reader, tomb.Context(nil), snapshot.Current, snapshot.Users, logf, opts)
+	if err != nil {
+		return err
+	}
+
+	raw, err := marshalSnapConfig(reader.Conf)
 	if err != nil {
 		backendRevert(restoreState)
 		return fmt.Errorf("cannot marshal saved config: %v", err)
@@ -295,7 +329,7 @@ func doRestore(task *state.Task, tomb *tomb.Tomb) error {
 	st.Lock()
 	defer st.Unlock()
 
-	if err := configSetSnapConfig(st, snapshot.Snap, (*json.RawMessage)(&buf)); err != nil {
+	if err := configSetSnapConfig(st, snapshot.Snap, raw); err != nil {
 		backendRevert(restoreState)
 		return fmt.Errorf("cannot set snap config: %v", err)
 	}
@@ -321,12 +355,12 @@ func undoRestore(task *state.Task, _ *tomb.Tomb) error {
 		return taskGetErrMsg(task, err, "snapshot")
 	}
 
-	buf, err := json.Marshal(restoreState.Config)
+	raw, err := marshalSnapConfig(restoreState.Config)
 	if err != nil {
 		return fmt.Errorf("cannot marshal saved config: %v", err)
 	}
 
-	if err := configSetSnapConfig(st, snapshot.Snap, (*json.RawMessage)(&buf)); err != nil {
+	if err := configSetSnapConfig(st, snapshot.Snap, raw); err != nil {
 		return fmt.Errorf("cannot restore saved config: %v", err)
 	}
 
@@ -414,7 +448,7 @@ func delayedCrossMgrInit() {
 	snapstate.EstimateSnapshotSize = EstimateSnapshotSize
 }
 
-func MockBackendSave(f func(context.Context, uint64, *snap.Info, map[string]interface{}, []string) (*client.Snapshot, error)) (restore func()) {
+func MockBackendSave(f func(context.Context, uint64, *snap.Info, map[string]interface{}, []string, *dirs.SnapDirOptions) (*client.Snapshot, error)) (restore func()) {
 	old := backendSave
 	backendSave = f
 	return func() {
