@@ -38,6 +38,7 @@ import (
 
 	"gopkg.in/check.v1"
 
+	"github.com/snapcore/snapd/asserts/snapasserts"
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil/sys"
@@ -54,19 +55,28 @@ import (
 	"github.com/snapcore/snapd/testutil"
 )
 
-type snapshotSuite struct{}
+type snapshotSuite struct {
+	testutil.BaseTest
+}
 
 var _ = check.Suite(&snapshotSuite{})
 
 // tie gocheck into testing
 func TestSnapshot(t *testing.T) { check.TestingT(t) }
 
-func (snapshotSuite) SetUpTest(c *check.C) {
+func (s *snapshotSuite) SetUpTest(c *check.C) {
+	s.BaseTest.SetUpTest(c)
 	dirs.SetRootDir(c.MkDir())
 	os.MkdirAll(dirs.SnapshotsDir, os.ModePerm)
+
+	restore := snapstate.MockEnforcedValidationSets(func(st *state.State) (*snapasserts.ValidationSets, error) {
+		return nil, nil
+	})
+	s.AddCleanup(restore)
 }
 
-func (snapshotSuite) TearDownTest(c *check.C) {
+func (s *snapshotSuite) TearDownTest(c *check.C) {
+	s.BaseTest.TearDownTest(c)
 	dirs.SetRootDir("/")
 }
 
@@ -730,7 +740,9 @@ exec /bin/tar "$@"
 	// task 1 (for "too-snap") will have errored
 	c.Check(tasks[1].Summary(), testutil.Contains, `"too-snap"`) // sanity check: task 1 is too-snap's
 	c.Check(tasks[1].Status(), check.Equals, state.ErrorStatus)
-	c.Check(strings.Join(tasks[1].Log(), "\n"), check.Matches, `\S+ ERROR cannot create archive: .* Permission denied .and \d+ more.`)
+	c.Check(strings.Join(tasks[1].Log(), "\n"), check.Matches, `(?ms)\S+ ERROR cannot create archive:
+/bin/tar: common/common-too-snap: .* Permission denied
+/bin/tar: Exiting with failure status due to previous errors`)
 
 	// task 2 (for "tri-snap") will have errored as well, hopefully, but it's a race (see the "tar" comment above)
 	c.Check(tasks[2].Summary(), testutil.Contains, `"tri-snap"`) // sanity check: task 2 is tri-snap's
@@ -744,6 +756,122 @@ exec /bin/tar "$@"
 	c.Check(string(out), check.Equals, "")
 }
 
+func (snapshotSuite) testSaveIntegrationTarFails(c *check.C, tarLogLines int, expectedErr string) {
+	if os.Geteuid() == 0 {
+		c.Skip("this test cannot run as root (runuser will fail)")
+	}
+	c.Assert(os.MkdirAll(dirs.SnapshotsDir, 0755), check.IsNil)
+	// sanity check: no files in snapshot dir
+	out, err := exec.Command("find", dirs.SnapshotsDir, "-type", "f").CombinedOutput()
+	c.Assert(err, check.IsNil)
+	c.Check(string(out), check.Equals, "")
+
+	homedir := filepath.Join(dirs.GlobalRootDir, "home", "a-user")
+	// mock tar so that it outputs a desired number of lines
+	tarFailScript := fmt.Sprintf(`
+export LANG=C
+for c in $(seq %d); do echo "log line $c" >&2 ; done
+exec /bin/tar "$@"
+`, tarLogLines)
+	if tarLogLines == 1 {
+		tarFailScript = "echo nope >&2 ; exit 1"
+	} else if tarLogLines == 0 {
+		tarFailScript = "exit 1"
+	}
+	mocktar := testutil.MockCommand(c, "tar", tarFailScript)
+	defer mocktar.Restore()
+
+	defer backend.MockUserLookup(func(username string) (*user.User, error) {
+		if username != "a-user" {
+			c.Fatalf("unexpected user %q", username)
+		}
+		return &user.User{
+			Uid:      fmt.Sprint(sys.Geteuid()),
+			Username: username,
+			HomeDir:  homedir,
+		}, nil
+	})()
+
+	o := overlord.Mock()
+	st := o.State()
+
+	stmgr, err := snapstate.Manager(st, o.TaskRunner())
+	c.Assert(err, check.IsNil)
+	o.AddManager(stmgr)
+	shmgr := snapshotstate.Manager(st, o.TaskRunner())
+	o.AddManager(shmgr)
+	o.AddManager(o.TaskRunner())
+
+	st.Lock()
+	defer st.Unlock()
+
+	sideInfo := &snap.SideInfo{RealName: "tar-fail-snap", Revision: snap.R(1)}
+	snapstate.Set(st, "tar-fail-snap", &snapstate.SnapState{
+		Active:   true,
+		Sequence: []*snap.SideInfo{sideInfo},
+		Current:  sideInfo.Revision,
+		SnapType: "app",
+	})
+	snaptest.MockSnap(c, "name: tar-fail-snap\nversion: v1", sideInfo)
+	c.Assert(os.MkdirAll(filepath.Join(homedir, "snap/tar-fail-snap/1/canary-tar-fail-snap"), 0755), check.IsNil)
+	c.Assert(os.MkdirAll(filepath.Join(homedir, "snap/tar-fail-snap/common"), 0755), check.IsNil)
+	// these dir permissions (000) make tar unhappy
+	c.Assert(os.Mkdir(filepath.Join(homedir, "snap/tar-fail-snap/common/common-tar-fail-snap"), 00), check.IsNil)
+
+	setID, saved, taskset, err := snapshotstate.Save(st, nil, []string{"a-user"})
+	c.Assert(err, check.IsNil)
+	c.Check(setID, check.Equals, uint64(1))
+	c.Check(saved, check.DeepEquals, []string{"tar-fail-snap"})
+
+	change := st.NewChange("save-snapshot", "...")
+	change.AddAll(taskset)
+
+	st.Unlock()
+	c.Assert(o.Settle(testutil.HostScaledTimeout(5*time.Second)), check.IsNil)
+	st.Lock()
+	c.Check(change.Err(), check.NotNil)
+	tasks := change.Tasks()
+	c.Assert(tasks, check.HasLen, 1)
+
+	// task 1 (for "too-snap") will have errored
+	c.Check(tasks[0].Summary(), testutil.Contains, `"tar-fail-snap"`) // sanity check: task 1 is too-snap's
+	c.Check(tasks[0].Status(), check.Equals, state.ErrorStatus)
+	c.Check(strings.Join(tasks[0].Log(), "\n"), check.Matches, expectedErr)
+
+	// no zips left behind, not for errors, not for undos \o/
+	out, err = exec.Command("find", dirs.SnapshotsDir, "-type", "f").CombinedOutput()
+	c.Assert(err, check.IsNil)
+	c.Check(string(out), check.Equals, "")
+}
+
+func (s *snapshotSuite) TestSaveIntegrationTarFailsManyLines(c *check.C) {
+	// cutoff at 5 lines, 3 lines of log + 2 lines from tar
+	s.testSaveIntegrationTarFails(c, 3, `(?ms)\S+ ERROR cannot create archive:
+log line 1
+log line 2
+log line 3
+/bin/tar: common/common-tar-fail-snap: .* Permission denied
+/bin/tar: Exiting with failure status due to previous errors`)
+}
+
+func (s *snapshotSuite) TestSaveIntegrationTarFailsTrimmedLines(c *check.C) {
+	s.testSaveIntegrationTarFails(c, 10, `(?ms)\S+ ERROR cannot create archive \(showing last 5 lines out of 12\):
+log line 8
+log line 9
+log line 10
+/bin/tar: common/common-tar-fail-snap: .* Permission denied
+/bin/tar: Exiting with failure status due to previous errors`)
+}
+
+func (s *snapshotSuite) TestSaveIntegrationTarFailsSingleLine(c *check.C) {
+	s.testSaveIntegrationTarFails(c, 1, `(?ms)\S+ ERROR cannot create archive:
+nope`)
+}
+
+func (s *snapshotSuite) TestSaveIntegrationTarFailsNoLines(c *check.C) {
+	s.testSaveIntegrationTarFails(c, 0, `(?ms)\S+ ERROR tar failed: exit status 1`)
+}
+
 func (snapshotSuite) TestRestoreChecksIterError(c *check.C) {
 	defer snapshotstate.MockBackendIter(func(context.Context, func(*backend.Reader) error) error {
 		return errors.New("bzzt")
@@ -1022,6 +1150,21 @@ func (snapshotSuite) TestRestore(c *check.C) {
 }
 
 func (snapshotSuite) TestRestoreIntegration(c *check.C) {
+	testRestoreIntegration(c, dirs.UserHomeSnapDir, nil)
+}
+
+func (snapshotSuite) TestRestoreIntegrationHiddenSnapDir(c *check.C) {
+	opts := &dirs.SnapDirOptions{HiddenSnapDataDir: true}
+
+	restore := snapstate.MockGetSnapDirOptions(func(*state.State) (*dirs.SnapDirOptions, error) {
+		return opts, nil
+	})
+	defer restore()
+
+	testRestoreIntegration(c, dirs.HiddenSnapDataHomeDir, opts)
+}
+
+func testRestoreIntegration(c *check.C, snapDataDir string, opts *dirs.SnapDirOptions) {
 	if os.Geteuid() == 0 {
 		c.Skip("this test cannot run as root (runuser will fail)")
 	}
@@ -1066,16 +1209,16 @@ func (snapshotSuite) TestRestoreIntegration(c *check.C) {
 		snapInfo := snaptest.MockSnap(c, fmt.Sprintf("{name: %s, version: v1}", name), sideInfo)
 
 		for _, home := range []string{homedirA, homedirB} {
-			c.Assert(os.MkdirAll(filepath.Join(home, "snap", name, fmt.Sprint(i+1), "canary-"+name), 0755), check.IsNil)
-			c.Assert(os.MkdirAll(filepath.Join(home, "snap", name, "common", "common-"+name), 0755), check.IsNil)
+			c.Assert(os.MkdirAll(filepath.Join(home, snapDataDir, name, fmt.Sprint(i+1), "canary-"+name), 0755), check.IsNil)
+			c.Assert(os.MkdirAll(filepath.Join(home, snapDataDir, name, "common", "common-"+name), 0755), check.IsNil)
 		}
 
-		_, err := backend.Save(context.TODO(), 42, snapInfo, nil, []string{"a-user", "b-user"})
+		_, err := backend.Save(context.TODO(), 42, snapInfo, nil, []string{"a-user", "b-user"}, opts)
 		c.Assert(err, check.IsNil)
 	}
 
 	// move the old away
-	c.Assert(os.Rename(filepath.Join(homedirA, "snap"), filepath.Join(homedirA, "snap.old")), check.IsNil)
+	c.Assert(os.Rename(filepath.Join(homedirA, snapDataDir), filepath.Join(homedirA, "snap.old")), check.IsNil)
 	// remove b-user's home
 	c.Assert(os.RemoveAll(homedirB), check.IsNil)
 
@@ -1099,8 +1242,8 @@ func (snapshotSuite) TestRestoreIntegration(c *check.C) {
 	}
 
 	// check it was all brought back \o/
-	out, err := exec.Command("diff", "-rN", filepath.Join(homedirA, "snap"), filepath.Join("snap.old")).CombinedOutput()
-	c.Assert(err, check.IsNil)
+	out, err := exec.Command("diff", "-rN", filepath.Join(homedirA, snapDataDir), filepath.Join("snap.old")).CombinedOutput()
+	c.Check(err, check.IsNil)
 	c.Check(string(out), check.Equals, "")
 }
 
@@ -1147,7 +1290,7 @@ func (snapshotSuite) TestRestoreIntegrationFails(c *check.C) {
 		c.Assert(os.MkdirAll(filepath.Join(homedir, "snap", name, fmt.Sprint(i+1), "canary-"+name), 0755), check.IsNil)
 		c.Assert(os.MkdirAll(filepath.Join(homedir, "snap", name, "common", "common-"+name), 0755), check.IsNil)
 
-		_, err := backend.Save(context.TODO(), 42, snapInfo, nil, []string{"a-user"})
+		_, err := backend.Save(context.TODO(), 42, snapInfo, nil, []string{"a-user"}, nil)
 		c.Assert(err, check.IsNil)
 	}
 
@@ -1719,7 +1862,7 @@ func (snapshotSuite) TestEstimateSnapshotSize(c *check.C) {
 		Current:  sideInfo.Revision,
 	})
 
-	defer snapshotstate.MockBackendEstimateSnapshotSize(func(info *snap.Info, users []string) (uint64, error) {
+	defer snapshotstate.MockBackendEstimateSnapshotSize(func(*snap.Info, []string, *dirs.SnapDirOptions) (uint64, error) {
 		return 123, nil
 	})()
 
@@ -1740,7 +1883,7 @@ func (snapshotSuite) TestEstimateSnapshotSizeWithConfig(c *check.C) {
 		Current:  sideInfo.Revision,
 	})
 
-	defer snapshotstate.MockBackendEstimateSnapshotSize(func(info *snap.Info, users []string) (uint64, error) {
+	defer snapshotstate.MockBackendEstimateSnapshotSize(func(*snap.Info, []string, *dirs.SnapDirOptions) (uint64, error) {
 		return 100, nil
 	})()
 
@@ -1768,7 +1911,7 @@ func (snapshotSuite) TestEstimateSnapshotSizeError(c *check.C) {
 		Current:  sideInfo.Revision,
 	})
 
-	defer snapshotstate.MockBackendEstimateSnapshotSize(func(info *snap.Info, users []string) (uint64, error) {
+	defer snapshotstate.MockBackendEstimateSnapshotSize(func(*snap.Info, []string, *dirs.SnapDirOptions) (uint64, error) {
 		return 0, fmt.Errorf("an error")
 	})()
 
@@ -1789,7 +1932,7 @@ func (snapshotSuite) TestEstimateSnapshotSizeWithUsers(c *check.C) {
 	})
 
 	var gotUsers []string
-	defer snapshotstate.MockBackendEstimateSnapshotSize(func(info *snap.Info, users []string) (uint64, error) {
+	defer snapshotstate.MockBackendEstimateSnapshotSize(func(info *snap.Info, users []string, opts *dirs.SnapDirOptions) (uint64, error) {
 		gotUsers = users
 		return 0, nil
 	})()
